@@ -21,5 +21,5 @@ from django.utils.translation import gettext_lazy as _
 
 
 class DjangoCAConfig(AppConfig):  # pylint: disable=missing-class-docstring
-    name = 'django_ca'
-    verbose_name = _('Certificate Authority')
+    name = "django_ca"
+    verbose_name = _("Certificate Authority")
