@@ -18,6 +18,7 @@
 
 import argparse
 import os
+import typing
 
 from cryptography import x509
 from cryptography.hazmat.backends import default_backend
@@ -27,40 +28,64 @@ from cryptography.hazmat.primitives.serialization import PrivateFormat
 
 from django.core.files.base import ContentFile
 from django.core.management.base import CommandError
+from django.core.management.base import CommandParser
 
 from ... import ca_settings
-from ...extensions import IssuerAlternativeName
 from ...models import CertificateAuthority
 from ...utils import ca_storage
+from ..actions import PasswordAction
 from ..base import BaseCommand
-from ..base import CertificateAuthorityDetailMixin
-from ..base import PasswordAction
+from ..mixins import CertificateAuthorityDetailMixin
 
 
-class Command(BaseCommand, CertificateAuthorityDetailMixin):  # pylint: disable=missing-class-docstring
+class Command(CertificateAuthorityDetailMixin, BaseCommand):
+    """Implement :command:`manage.py import_ca`."""
+
     help = """Import an existing certificate authority.
 
 Note that the private key will be copied to the directory configured by the CA_DIR setting."""
 
-    def add_arguments(self, parser):
-        self.add_ca(parser, '--parent',
-                    help='''Make the CA an intermediate CA of the named CA. By default, this is a
-                    new root CA.''', no_default=True)
+    def add_arguments(self, parser: CommandParser) -> None:
+        self.add_ca(
+            parser,
+            "--parent",
+            help_text="Make the CA an intermediate CA of the named CA. By default, this is a new root CA.",
+            no_default=True,
+        )
         self.add_password(
-            parser, help='Password used to encrypt the private key. Pass no argument to be prompted.')
-        parser.add_argument('--import-password', nargs='?', action=PasswordAction, metavar='PASSWORD',
-                            prompt='Password to import CA: ',
-                            help='Password for the private key.')
+            parser, help_text="Password used to encrypt the private key. Pass no argument to be prompted."
+        )
+        parser.add_argument(
+            "--import-password",
+            nargs="?",
+            action=PasswordAction,
+            metavar="PASSWORD",
+            prompt="Password to import CA: ",
+            help="Password for the private key.",
+        )
 
         self.add_ca_args(parser)
 
-        parser.add_argument('name', help='Human-readable name of the CA')
-        parser.add_argument('key', help='Path to the private key (PEM or DER format).',
-                            type=argparse.FileType('rb'))
-        parser.add_argument('pem', help='Path to the public key (PEM or DER format).',
-                            type=argparse.FileType('rb'))
+        parser.add_argument("name", help="Human-readable name of the CA")
+        parser.add_argument(
+            "key", help="Path to the private key (PEM or DER format).", type=argparse.FileType("rb")
+        )
+        parser.add_argument(
+            "pem", help="Path to the public key (PEM or DER format).", type=argparse.FileType("rb")
+        )
 
-    def handle(self, name, key, pem, **options):  # pylint: disable=arguments-differ
+    def handle(  # type: ignore[override]
+        self,
+        name: str,
+        key: typing.BinaryIO,
+        pem: typing.BinaryIO,
+        parent: typing.Optional[CertificateAuthority],
+        password: typing.Optional[bytes],
+        import_password: typing.Optional[bytes],
+        issuer_alternative_name: typing.Optional[str],
+        issuer_url: typing.Optional[str],
+        **options: typing.Any
+    ) -> None:
         if not os.path.exists(ca_settings.CA_DIR):
             try:
                 os.makedirs(ca_settings.CA_DIR)
@@ -68,24 +93,28 @@ Note that the private key will be copied to the directory configured by the CA_D
                 pem.close()
                 key.close()
                 raise CommandError(
-                    '%s: Could not create CA_DIR: Permission denied.' % ca_settings.CA_DIR) from ex
+                    "%s: Could not create CA_DIR: Permission denied." % ca_settings.CA_DIR
+                ) from ex
             # FileNotFoundError shouldn't happen, whole point of this block is to create it
 
-        import_password = options['import_password']
         pem_data = pem.read()
         key_data = key.read()
-        crl_url = '\n'.join(options['crl_url'])
+        crl_url = "\n".join(options["crl_url"])
 
         # close reader objects (otherwise we get a ResourceWarning)
         key.close()
         pem.close()
 
-        issuer_alternative_name = options[IssuerAlternativeName.key]
         if issuer_alternative_name is None:  # pragma: no branch - no CA sets this
-            issuer_alternative_name = ''
+            issuer_alternative_name = ""
 
-        ca = CertificateAuthority(name=name, parent=options['parent'], issuer_url=options['issuer_url'],
-                                  issuer_alt_name=issuer_alternative_name, crl_url=crl_url)
+        ca = CertificateAuthority(
+            name=name,
+            parent=parent,
+            issuer_url=issuer_url,
+            issuer_alt_name=issuer_alternative_name,
+            crl_url=crl_url,
+        )
 
         # load public key
         try:
@@ -94,9 +123,9 @@ Note that the private key will be copied to the directory configured by the CA_D
             try:
                 pem_loaded = x509.load_der_x509_certificate(pem_data, default_backend())
             except Exception as ex:
-                raise CommandError('Unable to load public key.') from ex
-        ca.x509 = pem_loaded
-        ca.private_key_path = ca_storage.generate_filename('%s.key' % ca.serial.replace(':', ''))
+                raise CommandError("Unable to load public key.") from ex
+        ca.update_certificate(pem_loaded)
+        ca.private_key_path = ca_storage.generate_filename("%s.key" % ca.serial.replace(":", ""))
 
         # load private key
         try:
@@ -105,22 +134,24 @@ Note that the private key will be copied to the directory configured by the CA_D
             try:
                 key_loaded = serialization.load_der_private_key(key_data, import_password, default_backend())
             except Exception as ex:
-                raise CommandError('Unable to load private key.') from ex
+                raise CommandError("Unable to load private key.") from ex
 
-        if options['password'] is None:
-            encryption = serialization.NoEncryption()
+        if password is None:
+            encryption: serialization.KeySerializationEncryption = serialization.NoEncryption()
         else:
-            encryption = serialization.BestAvailableEncryption(options['password'])
+            encryption = serialization.BestAvailableEncryption(password)
 
         # write private key to file
-        pem = key_loaded.private_bytes(encoding=Encoding.PEM, format=PrivateFormat.TraditionalOpenSSL,
-                                       encryption_algorithm=encryption)
+        pem_as_bytes = key_loaded.private_bytes(
+            encoding=Encoding.PEM, format=PrivateFormat.TraditionalOpenSSL, encryption_algorithm=encryption
+        )
 
         try:
-            ca_storage.save(ca.private_key_path, ContentFile(pem))
+            ca_storage.save(ca.private_key_path, ContentFile(pem_as_bytes))
         except PermissionError as ex:
             raise CommandError(
-                '%s: Permission denied: Could not open file for writing' % ca.private_key_path) from ex
+                "%s: Permission denied: Could not open file for writing" % ca.private_key_path
+            ) from ex
 
         # Only save CA to database if we loaded all data and copied private key
         ca.save()
