@@ -17,13 +17,16 @@
 """
 
 import argparse
+import typing
 
 from cryptography import x509
 from cryptography.hazmat.backends import default_backend
 
 from django.core.management.base import CommandError
+from django.core.management.base import CommandParser
 
 from ...models import Certificate
+from ...models import CertificateAuthority
 from ..base import BaseCommand
 
 
@@ -32,12 +35,15 @@ class Command(BaseCommand):  # pylint: disable=missing-class-docstring
 
 The authority that that signed the certificate must exist in the database."""
 
-    def add_arguments(self, parser):
+    def add_arguments(self, parser: CommandParser) -> None:
         self.add_ca(parser, allow_disabled=False)
-        parser.add_argument('pub', help='Path to the public key (PEM or DER format).',
-                            type=argparse.FileType('rb'))
+        parser.add_argument(
+            "pub", help="Path to the public key (PEM or DER format).", type=argparse.FileType("rb")
+        )
 
-    def handle(self, pub, **options):  # pylint: disable=arguments-differ
+    def handle(  # type: ignore[override]
+        self, pub: typing.BinaryIO, ca: CertificateAuthority, **options: typing.Any
+    ) -> None:
         pub_data = pub.read()
 
         # close reader objects (otherwise we get a ResourceWarning)
@@ -50,8 +56,8 @@ The authority that that signed the certificate must exist in the database."""
             try:
                 pub_loaded = x509.load_der_x509_certificate(pub_data, default_backend())
             except Exception as ex:
-                raise CommandError('Unable to load public key.') from ex
+                raise CommandError("Unable to load public key.") from ex
 
-        cert = Certificate(ca=options['ca'])
-        cert.x509 = pub_loaded
+        cert = Certificate(ca=ca)
+        cert.update_certificate(pub_loaded)
         cert.save()
