@@ -16,49 +16,64 @@
 .. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
 """
 
+import typing
+from datetime import timedelta
+
 from cryptography.hazmat.primitives.serialization import Encoding
 
 from django.core.management.base import CommandError
+from django.core.management.base import CommandParser
 
 from ... import ca_settings
 from ...extensions import ExtendedKeyUsage
 from ...extensions import KeyUsage
 from ...extensions import SubjectAlternativeName
 from ...extensions import TLSFeature
+from ...management.actions import CertificateAction
 from ...management.base import BaseSignCommand
-from ...management.base import CertificateAction
 from ...models import Certificate
+from ...models import CertificateAuthority
 from ...models import Watcher
+from ...subject import Subject
 
 
 class Command(BaseSignCommand):  # pylint: disable=missing-class-docstring
-    help = """Sign a CSR and output signed certificate. The defaults depend on the configured
-default profile, currently %s.""" % ca_settings.CA_DEFAULT_PROFILE
+    help = (
+        """Sign a CSR and output signed certificate. The defaults depend on the configured
+default profile, currently %s."""
+        % ca_settings.CA_DEFAULT_PROFILE
+    )
 
-    add_extensions_help = 'TODO'
-    subject_help = 'TODO'
+    add_extensions_help = "TODO"
+    subject_help = "TODO"
 
-    def add_arguments(self, parser):
+    def add_arguments(self, parser: CommandParser) -> None:
         self.add_base_args(parser, no_default_ca=True)
-        parser.add_argument('cert', action=CertificateAction, allow_revoked=True,
-                            help='The certificate to resign.')
+        parser.add_argument(
+            "cert", action=CertificateAction, allow_revoked=True, help="The certificate to resign."
+        )
 
-    def handle(self, *args, **options):  # pylint: disable=arguments-differ
-        cert = options['cert']
-        ca = options['ca']
+    def handle(  # type: ignore[override]
+        self,
+        cert: Certificate,
+        ca: typing.Optional[CertificateAuthority],
+        subject: typing.Optional[Subject],
+        expires: timedelta,
+        watch: typing.List[str],
+        password: typing.Optional[bytes],
+        **options: typing.Any
+    ) -> None:
         if not ca:
-            ca = options['ca'] = cert.ca
-        self.test_options(*args, **options)
+            ca = cert.ca
+        self.test_options(ca=ca, password=password, expires=expires, **options)
 
         # get list of watchers
-        if options['watch']:
-            watchers = [Watcher.from_addr(addr) for addr in options['watch']]
+        if watch:
+            watchers = [Watcher.from_addr(addr) for addr in watch]
         else:
             watchers = list(cert.watchers.all())
 
-        if options['subject']:
-            subject = options['subject']
-        else:
+        if subject is None:
             subject = cert.subject
 
         if not options[KeyUsage.key]:
@@ -82,43 +97,31 @@ default profile, currently %s.""" % ca_settings.CA_DEFAULT_PROFILE
             san = options[SubjectAlternativeName.key]
 
         kwargs = {
-            'subject': subject,
-            'password': options['password'],
-            'csr_format': Encoding.PEM,
-            'key_usage': key_usage,
-            'extended_key_usage': ext_key_usage,
-            'tls_feature': tls_feature,
-            'algorithm': options['algorithm'],
-            'expires': options['expires'],
-            'subject_alternative_name': san,
-            'cn_in_san': False,
-        }
-        kwargs = {
-            'algorithm': options['algorithm'],
-            'csr_format': Encoding.PEM,
-            'expires': options['expires'],
-            'extensions': [],
-            'password': options['password'],
-            'subject': subject,
-            'cn_in_san': False,  # we already copy the SAN/CN from the original cert
+            "algorithm": options["algorithm"],
+            "csr_format": Encoding.PEM,
+            "expires": expires,
+            "extensions": [],
+            "password": password,
+            "subject": subject,
+            "cn_in_san": False,  # we already copy the SAN/CN from the original cert
         }
 
         for ext in [key_usage, ext_key_usage, tls_feature, san]:
             if ext is not None:
-                kwargs['extensions'].append(ext)
+                kwargs["extensions"].append(ext)
 
-        if 'CN' not in kwargs['subject'] and not san:
+        if "CN" not in kwargs["subject"] and not san:
             raise CommandError("Must give at least a CN in --subject or one or more --alt arguments.")
 
         try:
-            cert = Certificate.objects.create_cert(ca=ca, csr=cert.csr, **kwargs)
+            cert = Certificate.objects.create_cert(ca=ca, csr=cert.csr.loaded, **kwargs)
         except Exception as ex:
             raise CommandError(ex) from ex
 
         cert.watchers.add(*watchers)
 
-        if options['out']:
-            with open(options['out'], 'w') as stream:
-                stream.write(cert.pub)
+        if options["out"]:
+            with open(options["out"], "w") as stream:
+                stream.write(cert.pub.pem)
         else:
-            self.stdout.write(cert.pub)
+            self.stdout.write(cert.pub.pem)
