@@ -17,6 +17,7 @@
 """
 
 import logging
+import typing
 from datetime import timedelta
 from http import HTTPStatus
 
@@ -35,28 +36,34 @@ from .profiles import profiles
 
 log = logging.getLogger(__name__)
 
+FuncTypeVar = typing.TypeVar("FuncTypeVar", bound=typing.Callable[..., typing.Any])
+
 try:
     from celery import shared_task
+    from celery.local import Proxy
 except ImportError:
-    def shared_task(func):
+
+    def shared_task(func: FuncTypeVar) -> "Proxy[FuncTypeVar]":
         """Dummy decorator so that we can use the decorator whether celery is installed or not."""
 
         # We do not yet need this, but might come in handy in the future:
-        #func.delay = lambda *a, **kw: func(*a, **kw)
-        #func.apply_async = lambda *a, **kw: func(*a, **kw)
-        return func
+        # func.delay = lambda *a, **kw: func(*a, **kw)
+        # func.apply_async = lambda *a, **kw: func(*a, **kw)
+        func.delay = func  # type: ignore[attr-defined]
+        return typing.cast("Proxy[FuncTypeVar]", func)
+
 
 # requests and josepy are optional dependencies for acme tasks
 try:
     import josepy as jose
     import requests
 except ImportError:  # pragma: no cover
-    jose = requests = None
+    jose = requests = None  # type: ignore[assignment]
 
 
-def run_task(task, *args, **kwargs):
+def run_task(task: "Proxy[FuncTypeVar]", *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
     """Function that passes `task` to celery or invokes it directly, depending on if Celery is installed."""
-    eager = kwargs.pop('eager', False)
+    eager = kwargs.pop("eager", False)
 
     if ca_settings.CA_USE_CELERY is True and eager is False:
         return task.delay(*args, **kwargs)
@@ -65,24 +72,29 @@ def run_task(task, *args, **kwargs):
 
 
 @shared_task
-def cache_crl(serial, **kwargs):
+def cache_crl(serial: str, **kwargs: typing.Any) -> None:
     """Task to cache the CRL for a given CA."""
     ca = CertificateAuthority.objects.get(serial=serial)
     ca.cache_crls(**kwargs)
 
 
 @shared_task
-def cache_crls(serials=None):
+def cache_crls(serials: typing.Optional[typing.Iterable[str]] = None) -> None:
     """Task to cache the CRLs for all CAs."""
+    if serials is None:  # pragma: no cover; just to make mypy happy
+        serials = []
+
     if not serials:
-        serials = CertificateAuthority.objects.usable().values_list('serial', flat=True)
+        serials = typing.cast(
+            typing.Iterable[str], CertificateAuthority.objects.usable().values_list("serial", flat=True)
+        )
 
     for serial in serials:
         run_task(cache_crl, serial)
 
 
 @shared_task
-def generate_ocsp_key(serial, **kwargs):
+def generate_ocsp_key(serial: str, **kwargs: typing.Any) -> typing.Tuple[str, str, int]:
     """Task to generate an OCSP key for the CA named by `serial`."""
     ca = CertificateAuthority.objects.get(serial=serial)
     private_path, cert_path, cert = ca.generate_ocsp_key(**kwargs)
@@ -90,72 +102,72 @@ def generate_ocsp_key(serial, **kwargs):
 
 
 @shared_task
-def generate_ocsp_keys(**kwargs):
+def generate_ocsp_keys(**kwargs: typing.Any) -> typing.List[typing.Tuple[str, str, int]]:
     """Task to generate an OCSP keys for all usable CAs."""
     keys = []
-    for serial in CertificateAuthority.objects.usable().values_list('serial', flat=True):
+    for serial in CertificateAuthority.objects.usable().values_list("serial", flat=True):
         keys.append(generate_ocsp_key(serial, **kwargs))
     return keys
 
 
 @shared_task
 @transaction.atomic
-def acme_validate_challenge(challenge_pk):
+def acme_validate_challenge(challenge_pk: int) -> None:
     """Validate an ACME challenge."""
     if not ca_settings.CA_ENABLE_ACME:
-        log.error('ACME is not enabled.')
+        log.error("ACME is not enabled.")
         return
 
     if jose is None:  # pragma: no cover
-        log.error('josepy is not installed, cannot do challenge validation.')
+        log.error("josepy is not installed, cannot do challenge validation.")
         return
 
     try:
         challenge = AcmeChallenge.objects.url().get(pk=challenge_pk)
     except AcmeChallenge.DoesNotExist:
-        log.error('Challenge with id=%s not found', challenge_pk)
+        log.error("Challenge with id=%s not found", challenge_pk)
         return
 
     # Whoever is invoking this task is responsible for setting the status to "processing" first.
     if challenge.status != AcmeChallenge.STATUS_PROCESSING:
-        log.error('%s: %s: Invalid state (must be %s)', challenge, challenge.status,
-                  AcmeChallenge.STATUS_PROCESSING)
+        log.error(
+            "%s: %s: Invalid state (must be %s)", challenge, challenge.status, AcmeChallenge.STATUS_PROCESSING
+        )
         return
 
     # If the auth cannot be used for validation, neither can this challenge. We check auth.usable instead of
     # challenge.usable b/c a challenge in the "processing" state is not "usable" (= it is already being used).
     if challenge.auth.usable is False:
-        log.error('%s: Authentication is not usable', challenge)
+        log.error("%s: Authentication is not usable", challenge)
         return
 
     # General data for challenge validation
     token = challenge.token
     value = challenge.auth.value
-    encoded = jose.encode_b64jose(token.encode('utf-8'))
+    encoded = jose.encode_b64jose(token.encode("utf-8"))
     thumbprint = challenge.auth.order.account.thumbprint
-    expected = f'{encoded}.{thumbprint}'
+    expected = f"{encoded}.{thumbprint}"
+
+    # Challenge is marked as invalid by default
+    challenge_valid = False
 
     if challenge.type == AcmeChallenge.TYPE_HTTP_01:
         if requests is None:  # pragma: no cover
-            log.error('requests is not installed, cannot do http-01 challenge validation.')
+            log.error("requests is not installed, cannot do http-01 challenge validation.")
             return
 
-        url = f'http://{value}/.well-known/acme-challenge/{encoded}'
+        url = f"http://{value}/.well-known/acme-challenge/{encoded}"
 
         # Validate HTTP challenge (only thing supported so far)
         try:
             response = requests.get(url, timeout=1)
 
             if response.status_code == HTTPStatus.OK:
-                received = response.text
-            else:
-                received = False
+                challenge_valid = response.text == expected
         except Exception as ex:  # pylint: disable=broad-except
             log.exception(ex)
-            received = False
     else:
         log.error("%s: Only HTTP-01 challenges supported so far", challenge)
-        received = False
 
     # Transition state of the challenge depending on if the challenge is valid or not. RFC8555, Section 7.1.6:
     #
@@ -173,7 +185,7 @@ def acme_validate_challenge(challenge_pk):
     #
     #   "* ready: The server agrees that the requirements have been fulfilled, and is awaiting finalization.
     #   Submit a finalization request."
-    if received == expected:
+    if challenge_valid:
         challenge.status = AcmeChallenge.STATUS_VALID
         challenge.validated = timezone.now()
         challenge.auth.status = AcmeAuthorization.STATUS_VALID
@@ -182,7 +194,7 @@ def acme_validate_challenge(challenge_pk):
         auths = AcmeAuthorization.objects.filter(order=challenge.auth.order)
         auths = auths.exclude(status=AcmeAuthorization.STATUS_VALID)
         if not auths.exclude(pk=challenge.auth.pk).exists():
-            log.info('Order is now valid')
+            log.info("Order is now valid")
             challenge.auth.order.status = AcmeOrder.STATUS_READY
     else:
         challenge.status = AcmeChallenge.STATUS_INVALID
@@ -198,7 +210,7 @@ def acme_validate_challenge(challenge_pk):
         #   If an error occurs at any of these stages, the order moves to the "invalid" state.
         challenge.auth.order.status = AcmeOrder.STATUS_INVALID
 
-    log.info('Challenge %s is %s', challenge.pk, challenge.status)
+    log.info("Challenge %s is %s", challenge.pk, challenge.status)
     challenge.save()
     challenge.auth.save()
     challenge.auth.order.save()
@@ -206,30 +218,28 @@ def acme_validate_challenge(challenge_pk):
 
 @shared_task
 @transaction.atomic
-def acme_issue_certificate(acme_certificate_pk):
+def acme_issue_certificate(acme_certificate_pk: int) -> None:
     """Actually issue an ACME certificate."""
     if not ca_settings.CA_ENABLE_ACME:
-        log.error('ACME is not enabled.')
+        log.error("ACME is not enabled.")
         return
 
     try:
-        acme_cert = AcmeCertificate.objects.select_related('order__account__ca').get(pk=acme_certificate_pk)
+        acme_cert = AcmeCertificate.objects.select_related("order__account__ca").get(pk=acme_certificate_pk)
     except AcmeCertificate.DoesNotExist:
-        log.error('Certificate with id=%s not found', acme_certificate_pk)
+        log.error("Certificate with id=%s not found", acme_certificate_pk)
         return
 
     if acme_cert.usable is False:
-        log.error('%s: Cannot issue certificate for this order', acme_cert.order)
+        log.error("%s: Cannot issue certificate for this order", acme_cert.order)
         return
 
     subject_alternative_names = [a.subject_alternative_name for a in acme_cert.order.authorizations.all()]
-    log.info('%s: Issuing certificate for %s', acme_cert.order, ',' .join(subject_alternative_names))
+    log.info("%s: Issuing certificate for %s", acme_cert.order, ",".join(subject_alternative_names))
 
-    extensions = {
-        SubjectAlternativeName.key: SubjectAlternativeName({'value': subject_alternative_names})
-    }
+    extensions = {SubjectAlternativeName.key: SubjectAlternativeName({"value": subject_alternative_names})}
 
-    profile = profiles['server']
+    profile = profiles["server"]
 
     # Honor not_after from the order if set
     if acme_cert.order.not_after:
@@ -241,7 +251,8 @@ def acme_issue_certificate(acme_certificate_pk):
 
     # Finally, actually create a certificate
     cert = Certificate.objects.create_cert(
-        acme_cert.order.account.ca, csr=csr, profile=profile, expires=expires, extensions=extensions)
+        acme_cert.order.account.ca, csr=csr, profile=profile, expires=expires, extensions=extensions
+    )
 
     acme_cert.cert = cert
     acme_cert.order.status = AcmeOrder.STATUS_VALID
@@ -251,12 +262,12 @@ def acme_issue_certificate(acme_certificate_pk):
 
 @shared_task
 @transaction.atomic
-def acme_cleanup():
+def acme_cleanup() -> None:
     """Cleanup expired ACME orders."""
 
     if not ca_settings.CA_ENABLE_ACME:
         # NOTE: Since this task does only cleanup, log message is only info.
-        log.info('ACME is not enabled, not doing anything.')
+        log.info("ACME is not enabled, not doing anything.")
         return
 
     # Delete orders that expired more then a day ago.
