@@ -11,16 +11,26 @@
 # You should have received a copy of the GNU General Public License along with django-ca.  If not,
 # see <http://www.gnu.org/licenses/>.
 
-"""Central functions to load CA key and cert as PKey/X509 objects."""
+"""Reusable utility functions used throughout django-ca."""
 
 import binascii
 import os
 import re
 import shlex
+import typing
+from collections import abc
 from datetime import datetime
 from datetime import timedelta
 from ipaddress import ip_address
 from ipaddress import ip_network
+from typing import Any
+from typing import Dict
+from typing import Iterable
+from typing import List
+from typing import Optional
+from typing import Tuple
+from typing import Type
+from typing import Union
 from urllib.parse import urlparse
 
 import idna
@@ -36,55 +46,74 @@ from cryptography.hazmat.primitives.serialization import Encoding
 from cryptography.x509.oid import NameOID
 
 from django.core.files.storage import get_storage_class
-from django.core.serializers.json import DjangoJSONEncoder
 from django.core.validators import URLValidator
 from django.utils.encoding import force_bytes
 from django.utils.encoding import force_str
-from django.utils.functional import Promise
 from django.utils.translation import gettext_lazy as _
 
 from . import ca_settings
+from .typehints import Expires
+from .typehints import Literal
+from .typehints import ParsableGeneralName
+from .typehints import ParsableGeneralNameList
+from .typehints import ParsableHash
+from .typehints import ParsableKeyCurve
+from .typehints import ParsableKeyType
+from .typehints import ParsableRelativeDistinguishedName
+from .typehints import PrivateKeyTypes
+from .typehints import SupportsIndex
 
 # List of possible subject fields, in order
-SUBJECT_FIELDS = ['C', 'ST', 'L', 'O', 'OU', 'CN', 'emailAddress', ]
+SUBJECT_FIELDS = [
+    "C",
+    "ST",
+    "L",
+    "O",
+    "OU",
+    "CN",
+    "emailAddress",
+]
 
 # Description strings for various X509 extensions, taken from "man x509v3_config".
-EXTENDED_KEY_USAGE_DESC = _('Purposes for which the certificate public key can be used for.')
-KEY_USAGE_DESC = _('Permitted key usages.')
+EXTENDED_KEY_USAGE_DESC = _("Purposes for which the certificate public key can be used for.")
+KEY_USAGE_DESC = _("Permitted key usages.")
 
 #: Regular expression to match RDNs out of a full x509 name.
-NAME_RE = re.compile(r'(?:/+|\A)\s*(?P<field>[^\s]*?)\s*'
-                     r'=(?P<quote>[\'"])?\s*(?P<content>(?(quote).*?|[^/]*))\s*'
-                     r'(?(quote)(?<!\\)(?P=quote))', re.I)
+NAME_RE = re.compile(
+    r"(?:/+|\A)\s*(?P<field>[^\s]*?)\s*"
+    r'=(?P<quote>[\'"])?\s*(?P<content>(?(quote).*?|[^/]*))\s*'
+    r"(?(quote)(?<!\\)(?P=quote))",
+    re.I,
+)
 
 #: Regular expression to match general names.
-GENERAL_NAME_RE = re.compile('^(email|URI|IP|DNS|RID|dirName|otherName):(.*)', flags=re.I)
+GENERAL_NAME_RE = re.compile("^(email|URI|IP|DNS|RID|dirName|otherName):(.*)", flags=re.I)
 
-#: Regular expression matching hexlified certificate serials
-SERIAL_RE = re.compile('^([0-9A-F][0-9A-F]:?)+[0-9A-F][0-9A-F]?$')
+#: Regular expression matching certificate serials as hex
+SERIAL_RE = re.compile("^([0-9A-F][0-9A-F]:?)+[0-9A-F][0-9A-F]?$")
 
 SAN_NAME_MAPPINGS = {
-    x509.DNSName: 'DNS',
-    x509.RFC822Name: 'email',
-    x509.DirectoryName: 'dirname',
-    x509.UniformResourceIdentifier: 'URI',
-    x509.IPAddress: 'IP',
-    x509.RegisteredID: 'RID',
-    x509.OtherName: 'otherName',
+    x509.DNSName: "DNS",
+    x509.RFC822Name: "email",
+    x509.DirectoryName: "dirname",
+    x509.UniformResourceIdentifier: "URI",
+    x509.IPAddress: "IP",
+    x509.RegisteredID: "RID",
+    x509.OtherName: "otherName",
 }
 
 #: Map OID objects to IDs used in subject strings
-OID_NAME_MAPPINGS = {
-    NameOID.COUNTRY_NAME: 'C',
-    NameOID.STATE_OR_PROVINCE_NAME: 'ST',
-    NameOID.LOCALITY_NAME: 'L',
-    NameOID.ORGANIZATION_NAME: 'O',
-    NameOID.ORGANIZATIONAL_UNIT_NAME: 'OU',
-    NameOID.COMMON_NAME: 'CN',
-    NameOID.EMAIL_ADDRESS: 'emailAddress',
-    NameOID.SERIAL_NUMBER: 'serialNumber',
-    NameOID.JURISDICTION_COUNTRY_NAME: 'jurisdictionCountryName',
-    NameOID.JURISDICTION_STATE_OR_PROVINCE_NAME: 'jurisdictionStateOrProvinceName',
+OID_NAME_MAPPINGS: Dict[x509.ObjectIdentifier, str] = {
+    NameOID.COUNTRY_NAME: "C",
+    NameOID.STATE_OR_PROVINCE_NAME: "ST",
+    NameOID.LOCALITY_NAME: "L",
+    NameOID.ORGANIZATION_NAME: "O",
+    NameOID.ORGANIZATIONAL_UNIT_NAME: "OU",
+    NameOID.COMMON_NAME: "CN",
+    NameOID.EMAIL_ADDRESS: "emailAddress",
+    NameOID.SERIAL_NUMBER: "serialNumber",
+    NameOID.JURISDICTION_COUNTRY_NAME: "jurisdictionCountryName",
+    NameOID.JURISDICTION_STATE_OR_PROVINCE_NAME: "jurisdictionStateOrProvinceName",
     NameOID.BUSINESS_CATEGORY: "businessCategory",
     NameOID.POSTAL_CODE: "postalCode",
     NameOID.STREET_ADDRESS: "streetAddress",
@@ -102,22 +131,38 @@ MULTIPLE_OIDS = (
 # uppercase values as keys for normalizing case
 NAME_CASE_MAPPINGS = {v.upper(): v for v in OID_NAME_MAPPINGS.values()}
 
+try:
+    # pylint: disable=unused-import,useless-import-alias
+    #         Import alias is for mypy (explicit re-export)
+    from django.utils.functional import classproperty as classproperty
+except ImportError:  # pragma: no cover
+    # NOTE: Official Django documentation states that this decorator is new in Django 3.1, but in reality
+    #       it is present (but undocumented) in Django 2.2 as well.
+    # Copy of classproperty from django 3.1 for older versions
+    # pylint: disable=invalid-name,missing-function-docstring
+    class classproperty:  # type: ignore
+        """
+        Decorator that converts a method with a single `cls` argument into a property
+        that can be accessed directly from the class.
+        """
 
-class LazyEncoder(DjangoJSONEncoder):
-    """Encoder that also encodes strings translated with gettext_lazy."""
+        def __init__(self, method=None):  # type: ignore
+            self.fget = method
 
-    def default(self, o):
-        if isinstance(o, Promise):
-            return force_str(o)
-        return super().default(o)
+        def __get__(self, instance, cls=None):  # type: ignore
+            return self.fget(cls)
+
+        def getter(self, method):  # type: ignore
+            self.fget = method
+            return self
 
 
-def sort_name(subject):
+def sort_name(subject: List[Tuple[str, str]]) -> List[Tuple[str, str]]:
     """Returns the subject in the correct order for a x509 subject."""
     return sorted(subject, key=lambda e: SUBJECT_FIELDS.index(e[0]))
 
 
-def encode_url(url):
+def encode_url(url: str) -> str:
     """IDNA encoding for domains in URLs.
 
     Examples::
@@ -130,15 +175,15 @@ def encode_url(url):
         'https://xn--exmple-cua.com:8000/foobar'
     """
     parsed = urlparse(url)
-    if parsed.port:
-        hostname = idna.encode(parsed.hostname).decode('utf-8')
-        parsed = parsed._replace(netloc='%s:%s' % (hostname, parsed.port))
+    if parsed.hostname and parsed.port:
+        hostname = idna.encode(parsed.hostname).decode("utf-8")
+        parsed = parsed._replace(netloc="%s:%s" % (hostname, parsed.port))
     else:
-        parsed = parsed._replace(netloc=idna.encode(parsed.netloc).decode('utf-8'))
+        parsed = parsed._replace(netloc=idna.encode(parsed.netloc).decode("utf-8"))
     return parsed.geturl()
 
 
-def encode_dns(name):
+def encode_dns(name: str) -> str:
     """IDNA encoding for domains.
 
     Examples::
@@ -152,14 +197,14 @@ def encode_dns(name):
         >>> encode_dns('*.exämple.com')
         '*.xn--exmple-cua.com'
     """
-    if name.startswith('*.'):
-        return '*.%s' % idna.encode(name[2:]).decode('utf-8')
-    if name.startswith('.'):
-        return '.%s' % idna.encode(name[1:]).decode('utf-8')
-    return idna.encode(name).decode('utf-8')
+    if name.startswith("*."):
+        return "*.%s" % idna.encode(name[2:]).decode("utf-8")
+    if name.startswith("."):
+        return ".%s" % idna.encode(name[1:]).decode("utf-8")
+    return idna.encode(name).decode("utf-8")
 
 
-def format_name(subject):
+def format_name(subject: Union[x509.Name, Iterable[Tuple[str, str]]]) -> str:
     """Convert a subject into the canonical form for distinguished names.
 
     This function does not take care of sorting the subject in any meaningful order.
@@ -174,10 +219,10 @@ def format_name(subject):
     if isinstance(subject, x509.Name):
         subject = [(OID_NAME_MAPPINGS[s.oid], s.value) for s in subject]
 
-    return '/%s' % ('/'.join(['%s=%s' % (force_str(k), force_str(v)) for k, v in subject]))
+    return "/%s" % ("/".join(["%s=%s" % (k, v) for k, v in subject]))
 
 
-def format_relative_name(name):
+def format_relative_name(name: Union[x509.RelativeDistinguishedName, Iterable[Tuple[str, str]]]) -> str:
     """Convert a relative name (RDN) into a canonical form.
 
     Examples::
@@ -192,10 +237,10 @@ def format_relative_name(name):
     if isinstance(name, x509.RelativeDistinguishedName):
         name = [(OID_NAME_MAPPINGS[s.oid], s.value) for s in name]
 
-    return '/%s' % ('/'.join(['%s=%s' % (force_str(k), force_str(v)) for k, v in name]))
+    return "/%s" % ("/".join(["%s=%s" % (k, v) for k, v in name]))
 
 
-def format_general_name(name):
+def format_general_name(name: x509.GeneralName) -> str:
     """Format a single general name.
 
     >>> import ipaddress
@@ -209,11 +254,11 @@ def format_general_name(name):
         value = format_name(name.value)
     else:
         value = name.value
-    return '%s:%s' % (SAN_NAME_MAPPINGS[type(name)], value)
+    return "%s:%s" % (SAN_NAME_MAPPINGS[type(name)], value)
 
 
-def is_power2(num):
-    """Return True if num is a power of 2.
+def is_power2(num: int) -> bool:
+    """Return True if `num` is a power of 2.
 
     >>> is_power2(4)
     True
@@ -223,7 +268,7 @@ def is_power2(num):
     return num != 0 and ((num & (num - 1)) == 0)
 
 
-def multiline_url_validator(value):
+def multiline_url_validator(value: str) -> None:
     """Validate that a TextField contains one valid URL per line.
 
     .. seealso:: https://docs.djangoproject.com/en/1.9/ref/validators/
@@ -234,7 +279,7 @@ def multiline_url_validator(value):
         validator(line)
 
 
-def add_colons(value, pad='0'):
+def add_colons(value: str, pad: str = "0") -> str:
     """Add colons after every second digit.
 
     This function is used in functions to prettify serials.
@@ -247,18 +292,18 @@ def add_colons(value, pad='0'):
 
     s : str
         The string to add colons to
-    pad : str, default
-        If not None, pad the string so that the last element always has two characters. The default is
-        ``"0"``.
+    pad : str, optional
+        If not an empty string, pad the string so that the last element always has two characters. The default
+        is ``"0"``.
     """
 
-    if len(value) % 2 == 1 and pad is not None:
-        value = '%s%s' % (pad, value)
+    if len(value) % 2 == 1 and pad:
+        value = "%s%s" % (pad, value)
 
-    return ':'.join([value[i:i + 2] for i in range(0, len(value), 2)])
+    return ":".join([value[i : i + 2] for i in range(0, len(value), 2)])
 
 
-def int_to_hex(i):
+def int_to_hex(i: int) -> str:
     """Create a hex-representation of the given serial.
 
     >>> int_to_hex(12345678)
@@ -267,16 +312,16 @@ def int_to_hex(i):
     return hex(i)[2:].upper()
 
 
-def bytes_to_hex(value):
+def bytes_to_hex(value: bytes) -> str:
     """Convert a bytes array to hex.
 
     >>> bytes_to_hex(b'test')
     '74:65:73:74'
     """
-    return add_colons(binascii.hexlify(value).upper().decode('utf-8'))
+    return add_colons(binascii.hexlify(value).upper().decode("utf-8"))
 
 
-def hex_to_bytes(value):
+def hex_to_bytes(value: str) -> bytes:
     """Convert a hex number to bytes.
 
     This should be the inverse of :py:func:`~django_ca.utils.bytes_to_hex`.
@@ -284,13 +329,13 @@ def hex_to_bytes(value):
     >>> hex_to_bytes('74:65:73:74')
     b'test'
     """
-    return binascii.unhexlify(value.replace(':', ''))
+    return binascii.unhexlify(value.replace(":", ""))
 
 
-def sanitize_serial(value):
+def sanitize_serial(value: str) -> str:
     """Sanitize a serial provided by user/untrusted input.
 
-    This function is intended to be used to get a serial as used internaly by **django-ca** from untrusted
+    This function is intended to be used to get a serial as used internally by **django-ca** from untrusted
     user input. Internally, serials are stored in upper case and without ``:`` and leading zeros, but user
     output adds at least ``:``.
 
@@ -300,15 +345,43 @@ def sanitize_serial(value):
         '1AB'
     """
 
-    serial = value.upper().replace(':', '')
-    if serial != '0':
-        serial = serial.lstrip('0')
-    if re.search('[^0-9A-F]', serial):
-        raise ValueError('%s: Serial has invalid characters' % value)
+    serial = value.upper().replace(":", "")
+    if serial != "0":
+        serial = serial.lstrip("0")
+    if re.search("[^0-9A-F]", serial):
+        raise ValueError("%s: Serial has invalid characters" % value)
     return serial
 
 
-def parse_name(name):
+def parse_csr(
+    csr: Union[x509.CertificateSigningRequest, str, bytes], csr_format: Encoding
+) -> x509.CertificateSigningRequest:
+    """Parse a CSR in the given format.
+
+    .. deprecated:: 1.18.0
+
+       This function is no longer useful and will be removed in django-ca 1.20.0.
+
+    Parameters
+    ----------
+
+    csr : str or bytes or :py:class:`~cg:cryptography.x509.CertificateSigningRequest`
+        The CSR to parse.
+    csr_format : :py:class:`~cg:cryptography.hazmat.primitives.serialization.Encoding`
+        The format that the CSR is in.
+    """
+
+    if isinstance(csr, x509.CertificateSigningRequest):  # pragma: no cover # not used since 1.18.0
+        return csr
+    if csr_format == Encoding.PEM:
+        return x509.load_pem_x509_csr(force_bytes(csr), default_backend())
+    if csr_format == Encoding.DER:
+        return x509.load_der_x509_csr(force_bytes(csr), default_backend())
+
+    raise ValueError("Unknown CSR format passed: %s" % csr_format)
+
+
+def parse_name(name: str) -> List[Tuple[str, str]]:
     """Parses a subject string as used in OpenSSLs command line utilities.
 
     The ``name`` is expected to be close to the subject format commonly used by OpenSSL, for example
@@ -357,7 +430,7 @@ def parse_name(name):
     try:
         items = [(NAME_CASE_MAPPINGS[t[0].upper()], force_str(t[2])) for t in NAME_RE.findall(name)]
     except KeyError as e:
-        raise ValueError('Unknown x509 name field: %s' % e.args[0]) from e
+        raise ValueError("Unknown x509 name field: %s" % e.args[0]) from e
 
     # Check that no OIDs not in MULTIPLE_OIDS occur more then once
     for key, oid in NAME_OID_MAPPINGS.items():
@@ -367,7 +440,7 @@ def parse_name(name):
     return sort_name(items)
 
 
-def x509_name(name):
+def x509_name(name: Union[List[Tuple[str, str]], str]) -> x509.Name:
     """Parses a subject into a :py:class:`x509.Name <cg:cryptography.x509.Name>`.
 
     If ``name`` is a string, :py:func:`parse_name` is used to parse it.
@@ -383,7 +456,7 @@ def x509_name(name):
     return x509.Name([x509.NameAttribute(NAME_OID_MAPPINGS[typ], force_str(value)) for typ, value in name])
 
 
-def x509_relative_name(name):
+def x509_relative_name(name: ParsableRelativeDistinguishedName) -> x509.RelativeDistinguishedName:
     """Parse a relative name (RDN) into a :py:class:`~cg:cryptography.x509.RelativeDistinguishedName`.
 
     >>> x509_relative_name('/CN=example.com')
@@ -396,12 +469,12 @@ def x509_relative_name(name):
     if isinstance(name, str):
         name = parse_name(name)
 
-    return x509.RelativeDistinguishedName([
-        x509.NameAttribute(NAME_OID_MAPPINGS[typ], force_str(value)) for typ, value in name
-    ])
+    return x509.RelativeDistinguishedName(
+        [x509.NameAttribute(NAME_OID_MAPPINGS[typ], force_str(value)) for typ, value in name]
+    )
 
 
-def validate_email(addr):
+def validate_email(addr: str) -> str:
     """Validate an email address.
 
     This function raises ``ValueError`` if the email address is not valid.
@@ -414,19 +487,19 @@ def validate_email(addr):
     ValueError: Invalid domain: bar com
 
     """
-    if '@' not in addr:
-        raise ValueError('Invalid email address: %s' % addr)
+    if "@" not in addr:
+        raise ValueError("Invalid email address: %s" % addr)
 
-    node, domain = addr.rsplit('@', 1)
+    node, domain = addr.rsplit("@", 1)
     try:
-        domain = idna.encode(domain).decode('utf-8')
-    except idna.core.IDNAError as e:
-        raise ValueError('Invalid domain: %s' % domain) from e
+        domain = idna.encode(domain).decode("utf-8")
+    except idna.IDNAError as e:
+        raise ValueError("Invalid domain: %s" % domain) from e
 
-    return '%s@%s' % (node, domain)
+    return "%s@%s" % (node, domain)
 
 
-def validate_hostname(hostname, allow_port=False):
+def validate_hostname(hostname: str, allow_port: bool = False) -> str:
     """Validate a hostname, optionally with a given port.
 
     >>> validate_hostname('example.com')
@@ -450,28 +523,55 @@ def validate_hostname(hostname, allow_port=False):
 
     """
     port = None
-    if allow_port is True and ':' in hostname:
-        hostname, port = hostname.rsplit(':', 1)
+    if allow_port is True and ":" in hostname:
+        hostname, port_str = hostname.rsplit(":", 1)
 
         try:
-            port = int(port)
+            port = int(port_str)
         except ValueError as e:
-            raise ValueError('%s: Port must be an integer' % port) from e
-        else:
-            if port < 1 or port > 65535:
-                raise ValueError('%s: Port must be between 1 and 65535' % port)
+            raise ValueError("%s: Port must be an integer" % port_str) from e
+
+        if port < 1 or port > 65535:
+            raise ValueError("%s: Port must be between 1 and 65535" % port)
 
     try:
-        encoded = idna.encode(hostname).decode('utf-8')
+        encoded: str = idna.encode(hostname).decode("utf-8")
     except idna.IDNAError as e:
-        raise ValueError('%s: Not a valid hostname' % hostname) from e
+        raise ValueError("%s: Not a valid hostname" % hostname) from e
 
     if allow_port is True and port is not None:
-        return '%s:%s' % (encoded, port)
+        return "%s:%s" % (encoded, port)
     return encoded
 
 
-def validate_key_parameters(key_size=None, key_type='RSA', ecc_curve=None):
+@typing.overload
+def validate_key_parameters(
+    key_size: Optional[int], key_type: Literal["DSA"], ecc_curve: ParsableKeyCurve = None
+) -> Tuple[int, Literal["DSA"], None]:
+    ...
+
+
+@typing.overload
+def validate_key_parameters(
+    key_size: Optional[int] = None,
+    key_type: Optional[Literal["RSA"]] = "RSA",
+    ecc_curve: ParsableKeyCurve = None,
+) -> Tuple[int, Literal["RSA"], None]:
+    ...
+
+
+@typing.overload
+def validate_key_parameters(
+    key_size: Optional[int], key_type: Literal["ECC"], ecc_curve: ParsableKeyCurve = None
+) -> Tuple[None, Literal["ECC"], ec.EllipticCurve]:
+    ...
+
+
+def validate_key_parameters(
+    key_size: Optional[int] = None,
+    key_type: Optional[ParsableKeyType] = "RSA",
+    ecc_curve: ParsableKeyCurve = None,
+) -> Tuple[Optional[int], Literal["RSA", "DSA", "ECC"], Optional[ec.EllipticCurve]]:
     """Validate parameters for private key generation and return sanitized values.
 
     This function can be used to fail early if invalid parameters are passed, before the private key is
@@ -487,27 +587,46 @@ def validate_key_parameters(key_size=None, key_type='RSA', ecc_curve=None):
     ValueError: 4000: Key size must be a power of two
     """
     if key_type is None:
-        key_type = 'RSA'
+        key_type = "RSA"
 
-    if key_type == 'ECC':
+    if key_type == "ECC":
         key_size = None
         ecc_curve = parse_key_curve(ecc_curve)
-    elif key_type in ['DSA', 'RSA']:
+    elif key_type in ["DSA", "RSA"]:
+        ecc_curve = None
         if key_size is None:
             key_size = ca_settings.CA_DEFAULT_KEY_SIZE
 
         if not is_power2(key_size):
             raise ValueError("%s: Key size must be a power of two" % key_size)
         if key_size < ca_settings.CA_MIN_KEY_SIZE:
-            raise ValueError("%s: Key size must be least %s bits" % (
-                key_size, ca_settings.CA_MIN_KEY_SIZE))
+            raise ValueError("%s: Key size must be least %s bits" % (key_size, ca_settings.CA_MIN_KEY_SIZE))
     else:
-        raise ValueError('%s: Unknown key type' % key_type)
+        raise ValueError("%s: Unknown key type" % key_type)
 
     return key_size, key_type, ecc_curve
 
 
-def generate_private_key(key_size, key_type, ecc_curve):
+@typing.overload
+def generate_private_key(key_size: int, key_type: Literal["DSA"], ecc_curve: None) -> dsa.DSAPrivateKey:
+    ...
+
+
+@typing.overload
+def generate_private_key(key_size: int, key_type: Literal["RSA"], ecc_curve: None) -> rsa.RSAPrivateKey:
+    ...
+
+
+@typing.overload
+def generate_private_key(
+    key_size: None, key_type: Literal["ECC"], ecc_curve: ec.EllipticCurve
+) -> ec.EllipticCurvePrivateKey:
+    ...
+
+
+def generate_private_key(
+    key_size: Optional[int], key_type: Literal["RSA", "DSA", "ECC"], ecc_curve: Optional[ec.EllipticCurve]
+) -> PrivateKeyTypes:
     """Generate a private key.
 
     This function assumes that you called :py:func:`~django_ca.utils.validate_key_parameters` on the input
@@ -529,18 +648,17 @@ def generate_private_key(key_size, key_type, ecc_curve):
     key
         A private key of the appropriate type.
     """
-    if key_type == 'DSA':
-        private_key = dsa.generate_private_key(key_size=key_size, backend=default_backend())
-    elif key_type == 'ECC':
-        private_key = ec.generate_private_key(ecc_curve, default_backend())
-    else:
-        private_key = rsa.generate_private_key(public_exponent=65537, key_size=key_size,
-                                               backend=default_backend())
+    if key_type == "DSA" and key_size is not None:
+        return dsa.generate_private_key(key_size=key_size, backend=default_backend())
+    if key_type == "ECC" and ecc_curve is not None:
+        return ec.generate_private_key(ecc_curve, default_backend())
+    if key_type == "RSA" and key_size is not None:
+        return rsa.generate_private_key(public_exponent=65537, key_size=key_size, backend=default_backend())
 
-    return private_key
+    raise ValueError("%s: Invalid key type." % key_type)
 
 
-def parse_general_name(name):
+def parse_general_name(name: ParsableGeneralName) -> x509.GeneralName:
     """Parse a general name from user input.
 
     This function will do its best to detect the intended type of any value passed to it:
@@ -600,8 +718,9 @@ def parse_general_name(name):
     if isinstance(name, x509.GeneralName):
         return name
     if not isinstance(name, str):
-        raise ValueError('Cannot parse general name %s: Must be of type str (was: %s).' %
-                         (name, type(name).__name__))
+        raise ValueError(
+            "Cannot parse general name %s: Must be of type str (was: %s)." % (name, type(name).__name__)
+        )
 
     typ = None
     match = GENERAL_NAME_RE.match(name)
@@ -610,19 +729,19 @@ def parse_general_name(name):
         typ = typ.lower()
 
     if typ is None:
-        if re.match('[a-z0-9]{2,}://', name):  # Looks like a URI
+        if re.match("[a-z0-9]{2,}://", name):  # Looks like a URI
             try:
                 return x509.UniformResourceIdentifier(encode_url(name))
             except idna.IDNAError:
                 pass
 
-        if '@' in name:  # Looks like an Email address
+        if "@" in name:  # Looks like an Email address
             try:
                 return x509.RFC822Name(validate_email(name))
             except ValueError:
                 pass
 
-        if name.strip().startswith('/'):  # maybe it's a dirname?
+        if name.strip().startswith("/"):  # maybe it's a dirname?
             return x509.DirectoryName(x509_name(name))
 
         # Try to parse this as IPAddress/Network
@@ -639,16 +758,16 @@ def parse_general_name(name):
         try:
             return x509.DNSName(encode_dns(name))
         except idna.IDNAError as e:
-            raise ValueError('Could not parse name: %s' % name) from e
+            raise ValueError("Could not parse name: %s" % name) from e
 
-    if typ == 'uri':
+    if typ == "uri":
         try:
             return x509.UniformResourceIdentifier(encode_url(name))
         except idna.IDNAError as e:
-            raise ValueError('Could not parse DNS name in URL: %s' % name) from e
-    elif typ == 'email':
+            raise ValueError("Could not parse DNS name in URL: %s" % name) from e
+    elif typ == "email":
         return x509.RFC822Name(validate_email(name))  # validate_email already raises ValueError
-    elif typ == 'ip':
+    elif typ == "ip":
         try:
             return x509.IPAddress(ip_address(name))
         except ValueError:
@@ -659,35 +778,35 @@ def parse_general_name(name):
         except ValueError:
             pass
 
-        raise ValueError('Could not parse IP address.')
-    elif typ == 'rid':
+        raise ValueError("Could not parse IP address.")
+    elif typ == "rid":
         return x509.RegisteredID(x509.ObjectIdentifier(name))
-    elif typ == 'othername':
-        regex = "(.*);(.*):(.*)"
-        if re.match(regex, name) is not None:
-            oid, asn_typ, val = re.match(regex, name).groups()
-            oid = x509.ObjectIdentifier(oid)
-            if asn_typ == 'UTF8':
-                val = val.encode('utf-8')
-            elif asn_typ == 'OctetString':
-                val = bytes(bytearray.fromhex(val))
-                val = OctetString(val).dump()
+    elif typ == "othername":
+        match = re.match("(.*);(.*):(.*)", name)
+        if match is not None:
+            oid, asn_typ, val = match.groups()
+            if asn_typ == "UTF8":
+                parsed_value = val.encode("utf-8")
+            elif asn_typ == "OctetString":
+                parsed_value = OctetString(bytes(bytearray.fromhex(val))).dump()
             else:
-                raise ValueError('Unsupported ASN type in otherName: %s' % asn_typ)
-            val = force_bytes(val)
-            return x509.OtherName(oid, val)
+                raise ValueError("Unsupported ASN type in otherName: %s" % asn_typ)
 
-        raise ValueError('Incorrect otherName format: %s' % name)
-    elif typ == 'dirname':
+            return x509.OtherName(x509.ObjectIdentifier(oid), parsed_value)
+
+        raise ValueError("Incorrect otherName format: %s" % name)
+    elif typ == "dirname":
         return x509.DirectoryName(x509_name(name))
     else:
         try:
             return x509.DNSName(encode_dns(name))
         except idna.IDNAError as e:
-            raise ValueError('Could not parse DNS name: %s' % name) from e
+            raise ValueError("Could not parse DNS name: %s" % name) from e
 
 
-def parse_hash_algorithm(value=None):
+def parse_hash_algorithm(
+    value: typing.Union[typing.Type[hashes.HashAlgorithm], ParsableHash] = None
+) -> hashes.HashAlgorithm:
     """Parse a hash algorithm value.
 
     The most common use case is to pass a str naming a class in
@@ -745,14 +864,15 @@ def parse_hash_algorithm(value=None):
         return value
     if isinstance(value, str):
         try:
-            return getattr(hashes, value.strip())()
+            algo: Type[hashes.HashAlgorithm] = getattr(hashes, value.strip())
+            return algo()
         except AttributeError as e:
-            raise ValueError('Unknown hash algorithm: %s' % value) from e
+            raise ValueError("Unknown hash algorithm: %s" % value) from e
 
-    raise ValueError('Unknown type passed: %s' % type(value).__name__)
+    raise ValueError("Unknown type passed: %s" % type(value).__name__)
 
 
-def parse_encoding(value=None):
+def parse_encoding(value: Optional[Union[str, Encoding]] = None) -> Encoding:
     """Parse a value to a valid encoding.
 
     This function accepts either a member of
@@ -772,18 +892,35 @@ def parse_encoding(value=None):
     if isinstance(value, Encoding):
         return value
     if isinstance(value, str):
-        if value == 'ASN1':
-            value = 'DER'
+        if value == "ASN1":
+            value = "DER"
 
         try:
-            return getattr(Encoding, value)
-        except AttributeError as e:
-            raise ValueError('Unknown encoding: %s' % value) from e
+            return Encoding[value]
+        except KeyError as e:
+            raise ValueError("Unknown encoding: %s" % value) from e
 
-    raise ValueError('Unknown type passed: %s' % type(value).__name__)
+    raise ValueError("Unknown type passed: %s" % type(value).__name__)
 
 
-def parse_key_curve(value=None):
+def parse_expires(expires: Expires = None) -> datetime:
+    """Parse a value specifying an expiry into a concrete datetime."""
+
+    now = datetime.utcnow().replace(second=0, microsecond=0)
+
+    if isinstance(expires, int):
+        return now + timedelta(days=expires)
+    if isinstance(expires, timedelta):
+        return now + expires
+    if isinstance(expires, datetime):
+        # NOTE: A datetime is passed when creating an intermediate CA and the expiry is limited by the expiry
+        # of the parent CA.
+        return expires.replace(second=0, microsecond=0)
+
+    return now + ca_settings.CA_DEFAULT_EXPIRES
+
+
+def parse_key_curve(value: ParsableKeyCurve = None) -> ec.EllipticCurve:
     """Parse an elliptic curve value.
 
     This function uses a value identifying an elliptic curve to return an
@@ -830,36 +967,35 @@ def parse_key_curve(value=None):
     if value is None:
         return ca_settings.CA_DEFAULT_ECC_CURVE
 
-    curve = getattr(ec, value.strip(), type)
+    curve: Type[ec.EllipticCurve] = getattr(ec, value.strip(), type)
     if not issubclass(curve, ec.EllipticCurve):
-        raise ValueError('%s: Not a known Eliptic Curve' % value)
+        raise ValueError("%s: Not a known Eliptic Curve" % value)
     return curve()
 
 
-def get_cert_builder(expires, serial=None):
-    """Get a basic X509 cert builder object.
-
-    .. TODO:: deprecate support for passing datetime as expires
+def get_cert_builder(expires: datetime, serial: Optional[int] = None) -> x509.CertificateBuilder:
+    """Get a basic X.509 certificate builder object.
 
     Parameters
     ----------
 
-    expires : datetime or timedelta
-        When this certificate will expire.
-    serial : int, optional
+    expires : datetime
         Serial number to set for this certificate. Use :py:func:`~cg:cryptography.x509.random_serial_number`
         to generate such a value. By default, a value will be generated.
     """
+
     now = datetime.utcnow().replace(second=0, microsecond=0)
 
+    # NOTE: Explicitly passing a serial is used when creating a CA, where we want to add extensions where the
+    # value references the serial.
     if serial is None:
         serial = x509.random_serial_number()
-    if expires is None:
-        expires = now + ca_settings.CA_DEFAULT_EXPIRES
-    elif isinstance(expires, timedelta):
-        expires = now + expires
-    else:
-        expires = expires.replace(second=0, microsecond=0)
+
+    if expires <= now:
+        raise ValueError("expires must be in the future")
+
+    # strip seconds and microseconds
+    expires = expires.replace(second=0, microsecond=0)
 
     builder = x509.CertificateBuilder()
     builder = builder.not_valid_before(now)
@@ -869,34 +1005,35 @@ def get_cert_builder(expires, serial=None):
     return builder
 
 
-def read_file(path):
+def read_file(path: str) -> bytes:
     """Read the file from the given path.
 
-    If ``path`` is an absolute path, reads a file from the local filesystem. For relative paths, read the file
-    using the storage backend configured using :ref:`CA_FILE_STORAGE <settings-ca-file-storage>`.
+    If ``path`` is an absolute path, reads a file from the local file system. For relative paths, read the
+    file using the storage backend configured using :ref:`CA_FILE_STORAGE <settings-ca-file-storage>`.
     """
     if os.path.isabs(path):
-        with open(path, 'rb') as stream:
-            return stream.read()
+        with open(path, "rb") as abs_stream:
+            return abs_stream.read()
 
     stream = ca_storage.open(path)
 
     try:
         # NOTE: In the python:3.9-rc-alpine3.10 Docker image, this is marked as a missed branch :-(
-        return stream.read()  # pragma: no branch
+        data: bytes = stream.read()  # pragma: no branch
+        return data
     finally:
         stream.close()
 
 
 # Note used currently, but left here for future reference
-#def write_private_file(path, data):
+# def write_private_file(path, data):
 #    """Function to write binary data to a file that will only be readable to the user."""
 #
 #    with os.fdopen(os.open(path, os.O_CREAT | os.O_WRONLY, 0o400), 'wb') as fh:
 #        fh.write(data)
 
 
-def shlex_split(val, sep):
+def shlex_split(val: str, sep: str) -> List[str]:
     """Split a character on the given set of characters.
 
     Example::
@@ -916,7 +1053,7 @@ def shlex_split(val, sep):
     return list(lex)
 
 
-class GeneralNameList(list):
+class GeneralNameList(List[x509.GeneralName]):
     """List that holds :py:class:`~cg:cryptography.x509.GeneralName` instances and parses ``str`` when added.
 
     A ``GeneralNameList`` is a ``list`` subclass that will always only hold
@@ -936,23 +1073,31 @@ class GeneralNameList(list):
         True
 
     """
-    def __init__(self, iterable=tuple()):
+
+    # pylint: disable=not-an-iterable; pylint does not detect generic iterables ("class Foo(List[Any])").
+    #                                  See also: https://github.com/PyCQA/pylint/issues/2568
+
+    def __init__(
+        self, iterable: Optional[Union[ParsableGeneralName, ParsableGeneralNameList]] = None
+    ) -> None:
+        if iterable is None:
+            iterable = []
         if isinstance(iterable, (str, x509.GeneralName)):
             iterable = [iterable]
 
         super().__init__(parse_general_name(v) for v in iterable)
 
-    def serialize(self):
+    def serialize(self) -> List[str]:
         """Generate a list of formatted names."""
-        for val in self:
-            yield format_general_name(val)
+        return [format_general_name(v) for v in self]
 
-    def __add__(self, value):  # self + other_list
-        if isinstance(value, GeneralNameList) is False:
+    def __add__(self, value: ParsableGeneralNameList) -> "GeneralNameList":
+        # self + other_list
+        if not isinstance(value, GeneralNameList):
             value = GeneralNameList(value)
         return GeneralNameList(list(self) + list(value))
 
-    def __contains__(self, value):  # value in self
+    def __contains__(self, value: Any) -> bool:  # value in self
         try:
             value = parse_general_name(value)
         except ValueError:
@@ -960,27 +1105,43 @@ class GeneralNameList(list):
 
         return list.__contains__(self, value)
 
-    def __eq__(self, other):  # value == other
+    def __eq__(self, other: Any) -> bool:  # value == other
         if isinstance(other, GeneralNameList) is False and isinstance(other, list) is True:
             other = GeneralNameList(other)
         return list.__eq__(self, other)
 
-    def __iadd__(self, value):  # self += value
+    def __iadd__(self, value: ParsableGeneralNameList) -> "GeneralNameList":  # self += value
         return list.__iadd__(self, (parse_general_name(v) for v in value))
 
-    def __repr__(self):
-        return '<GeneralNameList: %r>' % [format_general_name(v) for v in self]
+    def __repr__(self) -> str:
+        return "<GeneralNameList: %r>" % [format_general_name(v) for v in self]
 
-    def __setitem__(self, key, value):  # l[0] = 'example.com'
-        if isinstance(key, slice):  # l[0:1] = ['example.com']
+    @typing.overload
+    def __setitem__(self, key: SupportsIndex, value: ParsableGeneralName) -> None:  # pragma: no cover
+        ...
+
+    @typing.overload
+    def __setitem__(self, key: slice, value: ParsableGeneralNameList) -> None:  # pragma: no cover
+        ...
+
+    def __setitem__(
+        self, key: Union[SupportsIndex, slice], value: Union[ParsableGeneralNameList, ParsableGeneralName]
+    ) -> None:  # l[0] = 'example.com'
+        if isinstance(key, slice) and isinstance(value, abc.Iterable):
+            # equivalent to l[0:1] = ['example.com']
             list.__setitem__(self, key, (parse_general_name(v) for v in value))
-        else:
+        elif isinstance(key, int) and isinstance(value, (x509.GeneralName, str)):
+            # equivalent to l[0] = 'example.com'
             list.__setitem__(self, key, parse_general_name(value))
+        else:
+            raise TypeError("%s/%s: Invalid key/value type." % (key, value))
 
-    def append(self, o):
-        list.append(self, parse_general_name(o))
+    def append(self, value: ParsableGeneralName) -> None:
+        """Equivalent to list.append()."""
+        list.append(self, parse_general_name(value))
 
-    def count(self, value):
+    def count(self, value: ParsableGeneralName) -> int:
+        """Equivalent to list.count()."""
         try:
             value = parse_general_name(value)
         except ValueError:
@@ -988,23 +1149,34 @@ class GeneralNameList(list):
 
         return list.count(self, value)
 
-    def extend(self, iterable):
+    def extend(self, iterable: ParsableGeneralNameList) -> None:
+        """Equivalent to list.extend()."""
         list.extend(self, (parse_general_name(i) for i in iterable))
 
-    def index(self, value, *args):
+    def index(self, value: ParsableGeneralName, *args: int) -> int:
+        """Equivalent to list.index()."""
         return list.index(self, parse_general_name(value), *args)
 
-    def insert(self, index, o):
-        list.insert(self, index, parse_general_name(o))
+    def insert(self, index: int, value: ParsableGeneralName) -> None:
+        """Equivalent to list.insert()."""
+        list.insert(self, index, parse_general_name(value))
 
-    def remove(self, value):
+    def remove(self, value: ParsableGeneralName) -> None:
+        """Equivalent to list.remove()."""
         list.remove(self, parse_general_name(value))
 
 
-def get_crl_cache_key(serial, algorithm=hashes.SHA512, encoding=Encoding.DER, scope=None):
-    """Function to get a cache key for a CRL with the given parameters."""
-
-    return 'crl_%s_%s_%s_%s' % (serial, algorithm.name, encoding.name, scope)
+def get_crl_cache_key(
+    serial: str,
+    algorithm: hashes.HashAlgorithm = hashes.SHA512(),
+    encoding: Encoding = Encoding.DER,
+    scope: Optional[str] = None,
+) -> str:
+    """Get the cache key for a CRL with the given parameters."""
+    return "crl_%s_%s_%s_%s" % (serial, algorithm.name, encoding.name, scope)
 
 
-ca_storage = get_storage_class(ca_settings.CA_FILE_STORAGE)(**ca_settings.CA_FILE_STORAGE_KWARGS)
+# NOTE: get_storage_class is typed to Storage (but really returns the subclass FileSystemStorage).
+#       The default kwargs trigger a type error because the default works for the subclass.
+ca_storage_cls = get_storage_class(ca_settings.CA_FILE_STORAGE)
+ca_storage = ca_storage_cls(**ca_settings.CA_FILE_STORAGE_KWARGS)  # type: ignore
